"""eisenmp

A Python ``multiprocess, multi CPU`` module.
An example function cracks a game quest.

    ::

    Inheritance - proc: ProcEnv -> QueueCollect -> Mp
        Create Queue/Process -> Collect messages in boxes -> Manage, feed queues

    Thread names are in ProcEnv:
        QueueCollect [print_q, output_q, input_q, tools_q, info_q],
        GhettoGang [view_output_q_box, tools_q feeder],
        [ProcInfo]

"""
import time
import threading

import eisenmp.eisenmp_q_coll as coll
import eisenmp.eisenmp_procenv as procenv
import eisenmp.utils.eisenmp_utils as e_utils
import eisenmp.utils.eisenmp_constants as const
from eisenmp.eisenmp_q_coll import QueueCollect


class Mp(QueueCollect):
    """MultiProcessManager.

    """

    def __init__(self):
        super().__init__()
        self.kwargs = None

    def start(self, **kwargs):
        """enable Processes and eisenmp worker threads.
        """
        self.reset()
        self.kwargs = kwargs
        self.run_proc(**kwargs)

        self.enable_q_box_threads()  # [Baustelle] some q are collected in boxes, 'output', 'print' for later review

        self.enable_info_q()  # never disable, else sender blocks, nobody consumes from q
        if 'INFO_ENABLE' in kwargs and kwargs['INFO_ENABLE']:
            self.enable_info_thread()  # collect worker send nums from info box and shows % and ETA
        return

    def reset(self):
        """"""
        self.all_threads_stop = False  # frequent calls without exit, see bruteforce
        self.begin_proc_shutdown = False  # frequent calls without exit, see bruteforce
        if len(e_utils.Result.result_dict):
            e_utils.Result.result_dict = {}

    def run_q_feeder(self, **kwargs):
        """Threaded instance, run multiple q_feeder, called by manager of worker
        """
        self.kwargs.update(kwargs)  # upd boss kwargs with generator, queues and header_msg
        threading.Thread(name='eisenmp_q_feeder',  # better than class thread here, no overlap, interesting.
                         target=self.q_feeder,
                         ).start()

    def q_feeder(self):
        """Chunk list producer of generator input.

        - A ticket is attached as header to identify the workload (list chunks)
        - Serial number to rebuild the modified results in the right order
        """
        kw = self.kwargs
        generator = kw['generator']  # no generator for run_q_feeder, crash for sure
        num_rows = kw['NUM_ROWS'] if 'NUM_ROWS' in kw and kw['NUM_ROWS'] else const.NUM_ROWS  # processor workload
        feeder_input_q = kw['input_q'] if 'input_q' in kw else self.mp_input_q  # use default if not specified

        result_key = ''  # key name where Queue is stored as value, need a name to distinguish results in result dict
        for tup in self.q_name_id_lst:
            name, q_id, _ = tup  # name id q_ref
            if q_id == id(feeder_input_q):  # unique Python id for objects
                result_key = name
                break

        start = time.perf_counter()
        num_gen = e_utils.consecutive_number()
        while 1:
            if self.all_threads_stop:
                break
            chunk_lst = create_transport_ticket(num_gen, result_key)
            for _ in range(num_rows):
                try:
                    chunk_lst.append(next(generator))
                except StopIteration:
                    chunk_lst.append(const.STOP_MSG)  # signal stop to one worker module, worker module 'loader' to many
                    self.mp_print_q.put(f'\n\tgenerator empty, '
                                        f'run time iterator {round((time.perf_counter() - start))} seconds\n')
                    self.q_input_put(feeder_input_q, chunk_lst)
                    return

            self.q_input_put(feeder_input_q, chunk_lst)

    def q_input_put(self, feeder_input_q, chunk_lst):
        while 1:
            if self.all_threads_stop:
                break
            if feeder_input_q.empty():
                feeder_input_q.put(chunk_lst)
                break


def create_transport_ticket(num_gen, result_key):
    """Semicolon to split easy.
    """
    if not result_key:
        result_key = 'mp_input_q'
    serial_num = ';' + const.TICKET_ID_PREFIX + f'{str(next(num_gen))};'  # ';_TID_1;'
    return [result_key + serial_num]



"""Worker and mates module loader.
Mate worker modules MUST call a threaded start_function(),
else we hang. See watchdog.

"""

import os
import sys
import multiprocessing as mp

import eisenmp.utils.eisenmp_constants as const


class ToolBox:
    """Storage box for a Single Worker Process.
    Switch as much data fields to constant like shape. Used in worker.
    User can distinguish between custom and (automatic) in-build variables or constants.

    """
    def __init__(self):
        # default queues names from procenv
        self.mp_info_q = None  # performance data, or other
        self.mp_tools_q = None  # data too big to send with every list to worker
        self.mp_print_q = None  # formatted screen output with multiprocessor lock(), use sparse
        self.mp_input_q = None
        self.mp_output_q = None
        self.mp_process_q = None  # proc shutdown msg's
        # reserved names
        self.NEXT_LIST = None  # input_q, next list from your generator -> Ghetto is iterator, list creator
        self.WORKER_ID = None  # name id from Process name
        self.WORKER_PID = None  # process pid
        self.WORKER_NAME = None  # stop confirmations collected by GhettoGang and send shutdown signal to program
        self.MULTI_TOOL = None  # tools_q, can be any prerequisite object for a module
        self.STOP_MSG = None  # all q, not mp_print_q, if Boss StopIteration is raised, or this wrk informs other worker
        self.STOP_CONFIRM = ''  # output_q, GhettoGang collects msg to send shutdown signal to program
        self.OUTPUT_HEADER = ''  # identify proc result in output_q_box
        self.INPUT_HEADER = ''
        self.PERF_HEADER_ETA = None  # 'PROC_PERF_ETA_'
        self.PERF_CURRENT_ETA = None  # list rows done or other count
        self.kwargs = None


def module_path_load(file_path):
    """Imports the module from path and returns it in the env.
    """
    path, f_name = os.path.split(file_path)
    modulename, _ = os.path.splitext(f_name)

    if path not in sys.path:
        sys.path.insert(0, path)
    return __import__(modulename)


def all_worker_exit_msg(toolbox):
    """
    Warning: Signal stop event to [ALL] -----> worker MODULES, not to PROCESS.

    :params: toolbox: tools and Queues for processes
    """
    stop_token_lst = [const.OUTPUT_HEADER,
                      const.STOP_MSG]  # 'STOP' was sent if last list was produced; now we inform other worker

    for q in toolbox.ALL_QUEUES_LIST:  # next worker on any q reads 'stop_token_lst', except 'mp_print_q'
        if q.empty():
            q.put(stop_token_lst)

    toolbox.mp_output_q.put([toolbox.STOP_CONFIRM])  # essential msg for caller, count stop to exit
    toolbox.mp_print_q.put(f'\texit WORKER {toolbox.WORKER_ID}')
    pass


def mp_worker_entry(**kwargs):
    """Entry.
    We are 'disconnected' from parent process now.
    Only Queue communication. Threads can exec() 'string' commands.
    All references are dead. Variables ok. We read only, here.
    """
    worker = None
    # assembled vars and names
    name = mp.process.current_process().name
    proc_id = name.split('-')
    tool_box = {name: ToolBox()}
    tool_box[name].__dict__.update(kwargs)  # ToolBox class, add user defined attributes of ModuleConfiguration inst
    # defaults
    tool_box[name].WORKER_ID = int(proc_id[1])
    tool_box[name].WORKER_PID = int(os.getpid())
    tool_box[name].WORKER_NAME = name
    tool_box[name].STOP_MSG = const.STOP_MSG
    tool_box[name].STOP_CONFIRM = const.STOP_CONFIRM + name
    tool_box[name].OUTPUT_HEADER = const.OUTPUT_HEADER
    tool_box[name].PERF_HEADER_ETA = const.PERF_HEADER_ETA  # performance list header for ProcInfo
    tool_box[name].PERF_CURRENT_ETA = None
    tool_box[name].kwargs = kwargs

    toolbox = tool_box[name]  # use normal instance like

    # Module Loader
    mod_fun_lst = []
    for row in kwargs['worker_modules']:
        if len(row):
            path_t, ref_t = row.items()
            worker_path, worker_ref = path_t[1], ref_t[1]

            worker_mod = module_path_load(worker_path)  # str path to -> imported module now available
            fun_ref_exec = getattr(worker_mod, worker_ref)  # reference: can make function call now
            mod_fun_lst.append(fun_ref_exec)

    # Function executor
    if len(mod_fun_lst):
        mod_fun_len = len(mod_fun_lst)

        for workmate in range(mod_fun_len):
            if len(mod_fun_lst) >= 2:
                mate_fun = mod_fun_lst.pop()
                mate_fun(toolbox)
        worker = mod_fun_lst.pop()  # last but not least

    while 1:

        busy = worker(toolbox)  # until worker reads the iterator STOP msg
        if not busy:
            if 'STOP_MSG_DISABLE' not in toolbox.kwargs or not toolbox.kwargs['STOP_MSG_DISABLE']:
                all_worker_exit_msg(toolbox)  # stop msg in all queues, if not all loaded worker are threads
            break

    while 1:

        msg_lst = toolbox.mp_process_q.get()  # loader keeps proc alive and awaits, 'stop_proc'
        if const.STOP_PROCESS in msg_lst:

            for q in toolbox.ALL_QUEUES_LIST:  # multiple qs feeder may stick if first sent the worker stop
                if not q.empty():
                    q.get()
            break



""" eisenmp package constants definition
"""
GOT_RESULT_RESULT_THREAD = 'RESULT_INTERIM_GOT'  # append it to a list, use it for interim results during runtime
HEADER_MSG = 'INPUT_Q_HEAD'  # the default iterator writes a list header row, header and serial number, can be custom
NUM_ROWS = 1_000  # workload for one CPU core of generator output, the default iterator appends as row to a list
PERF_HEADER_ETA = 'PERF_HEADER_ETA'  # begin of performance list header to calc ETA
STOP_MSG = 'STOP'  # worker knows iterator is empty, no more new lists, can exit, return False
STOP_CONFIRM = 'WORKER_STOPS'  # worker writes stop message to mp_output_q
STOP_PROCESS = 'STOP_PROC'  # 'output_q_box_view' triggers all procs stop, if all worker confirmed stop
OUTPUT_HEADER = 'OUTPUT_HEADER'  # begin of output header, 'output_q_box_view' collects result lists
RESULTS_STORE = False
NUM_PROCS = None  # default, all processor cores
RESULT_LABEL = 'add a "RESULT_LABEL" var'
TICKET_ID_PREFIX = '_TID_'
ALL_QUEUES_LIST = 'ALL_QUEUES_LIST'  # module_loader puts stop msg in queues



"""ProcEnv

"""
import time
import multiprocessing as mp
from multiprocessing import Queue

import eisenmp.eisenmp_worker_loader as loader
import eisenmp.utils.eisenmp_utils as e_utils
import eisenmp.utils.eisenmp_constants as const


class ProcEnv:
    """Create the environment for worker processes on CPUs.
    All queues shared among processes.
    'maxsize=1' can be altered, should be tested and documented.

    - **Queues ONLY** in this version
    - custom Queue builder with a queue in a dict to show a name
    - another Queue builder can add a category name, dict in dict

    """

    def __init__(self):
        # CPU - process
        self.NUM_PROCS = self.core_count_get()  # user override in mod
        self.proc_list = []  # join processes at the end
        # Queues
        self.q_max_size = 1
        self.mp_info_q = Queue(maxsize=self.q_max_size)  # fake news and performance data
        self.mp_tools_q = Queue(maxsize=self.q_max_size)  # feeder is thread
        self.mp_print_q = Queue(maxsize=self.q_max_size)  # [!mp blocking!] OMG use sparse, formatted output
        self.mp_input_q = Queue(maxsize=self.q_max_size)  # order lists
        self.mp_output_q = Queue(maxsize=self.q_max_size)  # results and stop confirmation
        self.mp_process_q = Queue(maxsize=self.q_max_size)  # stop order
        self.queue_std_dict = {'mp_info_q': self.mp_info_q,
                               'mp_tools_q': self.mp_tools_q,
                               'mp_print_q': self.mp_print_q,
                               'mp_input_q': self.mp_input_q,
                               'mp_output_q': self.mp_output_q,
                               'mp_process_q': self.mp_process_q}
        self.queue_cust_dict_std = {}  # std dict, val is a q
        self.queue_cust_dict_cat = {}  # custom category, dict in dict
        self.q_lst = []  # all queues in a list, clean up; 'queue_lst_get'
        self.q_name_id_lst = [('mp_input_q (default)', id(self.mp_input_q), self.mp_info_q)]  # tuple (name, id, q_ref)

        # Threads - Collect queue grabber
        self.thread_list = []
        self.info_q_thread_name = 'eisenmp_info_q_thread'
        self.input_q_thread_name = 'eisenmp_input_q_thread'
        self.output_q_thread_name = 'eisenmp_output_q_thread'
        self.print_q_thread_name = 'eisenmp_print_q_thread'
        self.tools_q_thread_name = 'eisenmp_tools_q_thread'
        #    ProcInfo
        self.info_proc_info_thread = 'eisenmp_ProcInfo_thread'  # gang
        self.pi = None  # ProcInfo Sub-thread instance thread, start(), cancel()

        # Main switch
        self.all_threads_stop = False  # ends thread loops
        # update
        self.kwargs_env = {}

    def queue_cust_dict_std_create(self, *queue_name_maxsize: tuple):
        """create q, name and maxsize as unpacked list ('blue_q_7', 7)
        - Two queue creator functions. All use tuple to ease unpacking.

        'queue_cust_dict_std_create' - > 'queue_cust_dict_std'
        'queue_cust_dict_category_create' - > 'queue_cust_dict_cat'
        """
        for name, maxsize in queue_name_maxsize:
            self.queue_cust_dict_std[name] = Queue(maxsize=maxsize)
            self.q_name_id_lst.append((name, id(self.queue_cust_dict_std[name]), self.queue_cust_dict_std[name]))
            pass

    def queue_cust_dict_category_create(self, *queue_cat_name_maxsize: tuple):
        """('category_1', 'input_q_3', 10)
        """
        for cat, name, maxsize in queue_cat_name_maxsize:
            new_dct = {name: Queue(maxsize=maxsize)}
            if cat not in self.queue_cust_dict_cat:
                self.queue_cust_dict_cat[cat] = {}
            self.queue_cust_dict_cat[cat].update(new_dct)
            self.q_name_id_lst.append((cat + '|' + name, id(new_dct[name]), new_dct[name]))
            pass

    def queue_lst_get(self):
        """List of qs for shut down msg put in, of ...worker_loader.py
        """
        q_lst = [self.mp_input_q]

        for q in self.queue_cust_dict_std.values():  # custom, std dict
            self.q_lst.append(q)
        for cat_dct in self.queue_cust_dict_cat.values():  # custom, category dict
            for q in cat_dct.values():
                q_lst.append(q)

        self.q_lst.extend(q_lst)
        return self.q_lst

    @staticmethod
    def core_count_get():
        """"""
        num = 1 if not mp.cpu_count() else mp.cpu_count() / 2  # hyper thread
        return int(num)

    def kwargs_env_update_custom(self, **kwargs):
        """override default NUM_PROCS,
        'queue_lst_get' for worker loader stop msg in all qs
        """
        self.NUM_PROCS = kwargs['NUM_PROCS'] if 'NUM_PROCS' in kwargs and kwargs['NUM_PROCS'] else self.core_count_get()
        kwargs.update(self.queue_std_dict)
        kwargs.update(self.queue_cust_dict_std)
        kwargs.update(self.queue_cust_dict_cat)
        all_qs_dict = {const.ALL_QUEUES_LIST: self.queue_lst_get()}  # view q name,id,ref in debugger: 'q_name_id_lst'
        kwargs.update(all_qs_dict)
        return kwargs

    def run_proc(self, **kwargs):
        """Create a Process for each CPU core, if `num_proc` None set or not set.
        - kwargs dict is updated for worker 'toolbox', reveals all vars and dead references (spawn) available

        :params: kwargs: -
        :params: start_method: selection spawn, fork
        :params: START_SEQUENCE_NUM: useful if eisenmp instance is called often, process numbers rise, but start from 0
        :params: target=loader: the worker_loader module is loaded and keeps the process alive
        """
        kwargs = self.kwargs_env_update_custom(**kwargs)
        self.kwargs_env.update(kwargs)  # eat kwargs

        start_method = 'spawn' if 'START_METHOD' not in kwargs else kwargs['START_METHOD']
        mp.set_start_method(start_method, force=True)

        print(f'\nCreate {self.NUM_PROCS} processes.')
        for core in range(self.NUM_PROCS):
            print(core, end=" ")
            # start_sequence_num always starts zero and is independent of sub-process spawn number
            start_sequence_num = {'START_SEQUENCE_NUM': core}  # worker in proc can grab a specific queue 0=red,1=blue
            kwargs.update(start_sequence_num)

            proc = mp.Process(target=loader.mp_worker_entry,
                              kwargs=kwargs)
            proc.start()

            self.proc_list.append(proc)
        self.mp_print_q.put('\n')

    def stop_proc(self):
        """All worker must have confirmed shutdown msg."""
        for proc in range(len(self.proc_list)):
            self.mp_process_q.put(['Simon Says:', const.STOP_PROCESS])

        for process in self.proc_list:
            while process.is_alive():
                time.sleep(.1)

    def end_proc(self):
        """Graceful shutdown join.
        """
        for proc in self.proc_list:
            proc.join()
        print('\tProcesses are down.')

    def stop_thread(self):
        """"""
        for thread in self.thread_list:
            thread.cancel()

    def end_thread(self):
        """Instance and normal threads."""
        e_utils.thread_shutdown_wait(*self.thread_list)
        for t in self.thread_list:
            t.join()



# Configuration file for the Sphinx documentation builder.
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
#
# we choose Google Style Docstring Sections (see 20 lines below)
#
# def func(arg1, arg2):
#     """Summary line.
# 
#     Extended description of function.
# 
#     Args:
#         arg1 (int): Description of arg1
#         arg2 (str): Description of arg2
# 
#     Returns:
#         bool: Description of return value
# 
#     """
#     return True

# -- Project information -----------------------------------------------------
# !!! indentation of sphinx is mostly 3 leading spaces, code 4, block need one free line above and colon for head
# pip install -U sphinx
# pip install sphinx-rtd-theme
# pip install sphinxcontrib-napoleon
# Use sphinx-apidoc to build your API documentation:
# $ cd docs
# $ sphinx-apidoc -f -o source/ ../eisenradio/   (use full paths for both, __init__.py in the project folders)
# make html; .\make html, on powerShell

# Docstring Sections¶
# All of the following section headers are supported:
# Args (alias of Parameters)
# Arguments (alias of Parameters)
# Attention
# Attributes
# Caution
# Danger
# Error
# Example
# Examples
# Hint
# Important
# Keyword Args (alias of Keyword Arguments)
# Keyword Arguments
# Methods
# Note
# Notes
# Other Parameters
# Parameters
# Return (alias of Returns)
# Returns
# Raise (alias of Raises)
# Raises
# References
# See Also
# Tip
# Todo
# Warning
# Warnings (alias of Warning)
# Warn (alias of Warns)
# Warns
# Yield (alias of Yields)
# Yields

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'eisenmp'
copyright = 'BSD-3, 2023, René Horn'
author = 'René Horn'


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Add napoleon to the extensions list
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
pygments_style = 'sphinx'

html_static_path = ['_static']
html_logo = "eisenmp_logo.png"
html_logo_only = True
html_display_version = False
html_css_files = [
    "css-style.css",
]



import time
import threading

import eisenmp.utils.eisenmp_constants as constants


class ProcInfo(threading.Thread):
    """Thread; start(), cancel()
    Brute Force uses factorial as target to calculate.
    Prime uses 'range_num' of num generator as target.
        List rows done are the current state
    estimated time of arrival, ETA.
    So far.

    We sit on the QueueCollect and loop over the collected items in a list or dict (box).
    instantiated by GhettoGang

    """

    def __init__(self, name, print_q, info_q_box, **kwargs):
        super().__init__()
        # thread
        self.name = name
        self.daemon = True
        self.cancelled = False
        # info
        self.perf_dict_eta = {}
        self.info_shutdown = False
        self.print_q = print_q
        self.info_q_box = info_q_box  # dict
        self.INFO_THREAD_MAX = kwargs['INFO_THREAD_MAX'] if 'INFO_THREAD_MAX' in kwargs else None
        self.info_td_exec = kwargs['info_td_exec'] if 'info_td_exec' in kwargs else None

    def run(self):
        """feed init args instead of method args"""
        self.performance_coll()

    def cancel(self):
        self.info_shutdown = True
        self.cancelled = True

    def arrival_eta(self):
        """Calc ETA and percent done.
        """
        target = self.perf_dict_eta['target']
        current = self.perf_dict_eta['rows_done']
        start_time = self.perf_dict_eta['proc_start']

        if target and current:
            this_percent = current / target
            this_time = round((time.perf_counter() - start_time))

            est_arrival = this_time / this_percent
            time_left = est_arrival - this_time
            show_time = round(time_left, 1)
            show_percent = round((this_percent * 100), 1)
            return show_percent, show_time

    def perf_count_eta(self, info_box):
        """
        """
        for idx in range(len(info_box)):
            if idx in info_box.keys():
                if type(info_box[idx]) is list:

                    list_header = info_box[idx][0]
                    if list_header[:14] == constants.PERF_HEADER_ETA:
                        a_num = int(info_box[idx][1])
                        if self.INFO_THREAD_MAX:
                            self.perf_dict_eta['target'] = self.INFO_THREAD_MAX
                            if self.info_td_exec:
                                exec(self.info_td_exec)
                            else:
                                self.perf_dict_eta['rows_done'] += a_num  # ETA
                        info_box[idx] = None

    def perf_count_print_eta(self):
        """
        """
        if self.arrival_eta():
            percent, sec_left = self.arrival_eta()
            if 3_600 <= sec_left <= 86_400:
                msg_time = f' "{round(sec_left / 3_600, 1)}" hours left'
            elif sec_left >= 86_400:
                msg_time = f' "{round(sec_left / 86_400, 1)}" days left'
            else:
                msg_time = f' "{round(sec_left, 1)}" seconds left'

            msg = f'\n\t[ "{percent}%" done, {msg_time} ]\n'
            self.print_q.put(msg)

    def performance_coll(self):
        """Print time and percent.
        Loops forever if no search str or performance info.
        """
        self.perf_dict_eta = {'target': 0,
                              'rows_done': 0,
                              'proc_start': time.perf_counter()}
        info_box = self.info_q_box
        while not self.info_shutdown:  # self.perf_dict_eta['rows_done']
            self.perf_count_eta(info_box)
            self.perf_count_print_eta()
            for _ in range(40):
                if self.info_shutdown:
                    return
                time.sleep(.1)


if __name__ == '__main__':
    pass
    # ProcInfo()



import time
import threading


class Result:
    """Finest results only here.

    """
    result_dict = {}

    def result_dict_update(self, key: str, id_result: tuple) -> None:
        """Sort results yourself.
        If something is wrong, you see one TID is not there.

        Tuple for simple extraction: id, result = result_dict['PRIME_NUM']
        {'PRIME_NUM': [(ticket_id, result_lst), (ticket_id, result_lst)]}

        :params: key: str name of the queue feeder 'header_msg' argument
        :params: result_value_t: tuple(_TID_0, result)
        """
        if key not in self.result_dict:
            self.result_dict[key] = []
        self.result_dict[key].append(id_result)


def consecutive_number():
    """Want a stamp on each list header.
    Used for Queue messages get() and put in box_dict[num] = msg
    Can rebuild original order if worker puts result in a list
    with same num as order list.
    """
    idx = 0
    while 1:
        yield idx
        idx += 1


def thread_shutdown_wait(*threads):
    """We return if none of the thread names are listed anymore.
    Blocks!

    :params: *threads: arbitrary list of thread names
    """
    busy = True
    while busy:
        names_list = [thread.name for thread in threading.enumerate()]
        busy = True if any([True for thread in threads if thread in names_list]) else False
        time.sleep(.1)



import threading
import contextlib
import multiprocessing

import eisenmp.utils.eisenmp_utils as e_utils
import eisenmp.utils.eisenmp_constants as const
from eisenmp.eisenmp_procenv import ProcEnv
from eisenmp.utils.eisenmp_info import ProcInfo


class FunThread(threading.Thread):
    """Thread maker.

    """
    def __init__(self, name, fun_ref, *args, **kwargs):
        super().__init__()
        # thread
        self.name = name
        self.daemon = True
        self.cancelled = False
        # stuff
        self.fun_ref = fun_ref  # no ()
        self.args = args
        self.kwargs = kwargs

    def run(self):
        """_inst.start()"""
        self.fun_ref(*self.args, **self.kwargs)
        self.cancel()

    def cancel(self):
        self.cancelled = True


class QueueCollect(ProcEnv):
    """Queue message collector and printer. Logging from prints.
    Messages in input and output Q have header.
    One can decide later what to do. Alter, or monitor.

    - Output `can` be stored in a box. `store_result` set

    """

    def __init__(self):
        super().__init__()
        # collect box
        self.info_q_box = {}
        self.print_q_box = []
        self.output_q_box = {}  # dict multi thread access
        # internal lists
        self.stop_list = []  # proc stop answered, worker names collector
        self.result_lst = []  # collected findings from procs for this run
        # collect results [baustelle]
        self.e_utilsResult = e_utils.Result()  # interim result calc for long-running task

        self.begin_proc_shutdown = False

    def enable_q_box_threads(self):
        """Collect Q messages and put em in a box for review, if enabled
        """
        self.enable_print_q()
        self.enable_output_q()

    def enable_info_q(self):
        """Thread for loop."""
        infoQThread = FunThread('eisenmp_InfoQThread', self.info_q_loop)
        infoQThread.start()
        self.thread_list.append(infoQThread)

    def enable_info_thread(self):
        """Shows % done, time left, if fed with an end value"""
        args_inf = [self.info_proc_info_thread,
                    self.mp_print_q,
                    self.info_q_box]
        self.pi = ProcInfo(*args_inf, **self.kwargs_env)  # ProcInfo sits on a subclassed thread
        self.pi.start()  # cancel() the thread in 'thread_end_join'

    def enable_print_q(self):
        """Thread for loop."""
        printQThread = FunThread('eisenmp_PrintQThread', self.print_q_loop)
        printQThread.start()
        self.thread_list.append(printQThread)

    def enable_output_q(self):
        """Start a thread loop to not block the show.
        Want collect stop confirm worker msg and results, all lists
        """
        outputQThread = FunThread('eisenmp_OutputQThread', self.output_q_loop)
        outputQThread.start()
        # self.thread_list.append(outputQThread)

    def print_q_loop(self):
        """Use a Print Q and a thread for formatted printing.

        Use it only sparingly. BLOCKS the whole multiprocessing.
        """
        while 1:
            if self.all_threads_stop:
                break
            try:
                if not self.mp_print_q.empty():
                    with multiprocessing.Lock():
                        msg = self.mp_print_q.get()
                        self.print_q_box.append(msg)
                        print(msg)
            except Exception as e:
                with contextlib.redirect_stdout(None):
                    print(e)

    def info_q_loop(self):
        """Print info or collect statistics from boxed messages.
        Box is a standard dict with num generator for unique keys.
        """
        while 1:
            if self.all_threads_stop:
                break
            generator = e_utils.consecutive_number()
            while not self.all_threads_stop:
                if not self.mp_info_q.empty():
                    msg = self.mp_info_q.get()
                    num = next(generator)
                    self.info_q_box[num] = msg
                    pass

    def output_q_loop(self):
        """Grab output from Queue and put it in a box.
        Use consecutive_number to create unique keys.

        Note: Multiple threads can loop over the box (dict).
        """
        generator = e_utils.consecutive_number()
        while 1:
            if self.all_threads_stop:
                break
            if not self.mp_output_q.empty():
                worker_output = self.mp_output_q.get()
                serial_num = next(generator)
                self.output_q_box[serial_num] = worker_output
                is_STOP_MSG = self.output_q_search_stop_confirm(serial_num)
                self.output_q_box_view_results(serial_num) if not is_STOP_MSG else None

    def output_q_search_stop_confirm(self, serial_num):
        """Search stop msg of workers and put 'em in a list.
        Stop processes and threads, if list is full.
        {1: ['RESULT_HEADER;PRIME_NUM;_TID_1;Process-1', ['10000079']], 24: ['PROC_STOP;Process-5']}

        :params: serial_num: serial number of `output_q_loop`
        """
        outbox = self.output_q_box
        if type(outbox[serial_num]) is list:
            outbox_list = outbox[serial_num]

            list_header = outbox_list[0]
            if list_header[:len(const.STOP_CONFIRM)] == const.STOP_CONFIRM:
                if self.worker_mods_down_ask(list_header):
                    self.print_findings()

                    self.begin_proc_shutdown = True
                    self.all_threads_stop = True

                    self.stop_proc()
                    self.end_proc()
                    self.stop_thread()
                    self.end_thread()

                return True

    def output_q_box_view_results(self, serial_num):
        """Only `lists` with header accepted.
        Box is a dictionary. {key: val} 'output_q_loop()' -> output_q_box[num]: payload or stop msg

        {1: ['RESULT_HEADER;PRIME_NUM;_TID_1;Process-1', ['10000079']], 24: ['PROC_STOP;Process-5']}

        Result header: RESULT_HEADER
        Delivery header: PRIME_NUM;_TID_;0  ['header_msg' and eisenmp, custom iterator, loop ticket id, split(;)]

        Add your custom header_msg, taken from double Q Example:
        - mP.run_q_feeder(generator=generator_aud, feeder_input_q=audio_q_b1, header_msg='BATCH_1_A')  # custom head
        - mP.run_q_feeder(generator=generator_vid, feeder_input_q=video_q_b1, header_msg='BATCH_1_V')

        :params: serial_num: serial number of `output_q_loop`
        """
        outbox = self.output_q_box
        if type(outbox[serial_num]) is list:
            outbox_list = outbox[serial_num]  # value of consecutive_number key is the list with results or stop msg

            list_header = outbox_list[0]
            if list_header[:len(const.OUTPUT_HEADER)] == const.OUTPUT_HEADER:
                result_row = outbox_list[1]

                if 'RESULTS_STORE' in self.kwargs_env and self.kwargs_env['RESULTS_STORE']:  # store results switch
                    queue_header_msg, queue_ticket_id = self.proc_result_store(list_header)
                    if not queue_ticket_id:
                        return

                    res_val = (queue_ticket_id, result_row)  # tuple
                    self.e_utilsResult.result_dict_update(queue_header_msg, res_val)  # result_dict['PRIME_NUM']=res_val

                if 'RESULTS_PRINT' in self.kwargs_env and self.kwargs_env['RESULTS_PRINT']:
                    self.proc_result_list_findings(result_row)

    @staticmethod
    def proc_result_store(list_header):
        """"""
        queue_head = list_header[len(const.OUTPUT_HEADER):]  # remove 'RESULT_HEADER'
        queue_head_lst = queue_head.split(';')  # search ticket number
        queue_header_msg = queue_head_lst[0]  # 'PRIME_NUM', use as key for result dict
        queue_ticket_id = None
        for str_section in queue_head_lst:
            if str_section[:len(const.TICKET_ID_PREFIX)] == const.TICKET_ID_PREFIX:
                queue_ticket_id = str_section
                break
        return queue_header_msg, queue_ticket_id

    def proc_result_list_findings(self, p_result_row):
        """RESULT DICT in utils, collect all results
        Append to result list for print out at finish.
        Add to a dict to monitor results via additional thread during runtime. [baustelle]

        STOP msg was also appended!

        :params: p_result_row: process loop result list row, can be a list in this row
        """
        if const.STOP_MSG in str(p_result_row):
            return

        if type(p_result_row) is list:
            [self.result_lst.append(str(row) + '\n') for row in p_result_row if str(row)]  # internal list
        else:
            self.result_lst.append(str(p_result_row) + '\n')

    def worker_mods_down_ask(self, list_header):
        """All worker MODULES confirm shutdown.

        Worker is using the original name of its process in shut down msg.
        Process IS still running. Worker module entry function returned False.

        :params: list_header: answer of WORKER MODULE to stop request; string with proc id at end
        :returns: None; True if done
        """
        WORKER_NAME = list_header[len(const.STOP_CONFIRM):]
        self.stop_list.append(WORKER_NAME)
        pending = any([True for proc in self.proc_list if proc.name not in self.stop_list])
        if not pending:
            return True

    def print_findings(self):
        """Condensed result list for this run.
        A thread can sum the results in 'Result' class.
        """
        c_lst = []
        generator = (line.split('...') for line in list(set(self.result_lst)))
        try:
            c_lst = list(set([tup[2] for tup in generator]))
        except IndexError:
            c_lst.extend(self.result_lst)

        lbl = self.kwargs_env['RESULT_LABEL'] if 'RESULT_LABEL' in self.kwargs_env else const.RESULT_LABEL
        c_lst.insert(0, f'\n--- Result for [{lbl}]---\n')
        c_lst.append('    --- END ---\n')
        print('\n'.join(c_lst))  # procs down

        if 'RESULTS_DICT_PRINT' in self.kwargs_env and self.kwargs_env['RESULTS_DICT_PRINT']:
            print(e_utils.Result.result_dict)
